from math import floor
from typing import List, Literal, Tuple

import numpy as np
from scipy.spatial.distance import cdist
from sklearn.neighbors import KNeighborsClassifier


class CDistAndKNN:
    """
    This algorithm first counts the exact result for the part of ally_villages
    using `cdist` and then with `KNN model` with `N=3` tries to predict the rest,
    see `result` method.

    NOTES
    -----
    With `precision=1` and `_max_ally`, `_max_enemy` equals to something
    very big like `10**10`, it is the brute force.
    """

    def __init__(
        self,
        ally_villages: np.ndarray,
        enemy_villages: np.ndarray,
        min_radius: float,
        max_radius: float,
        _precision: float = 0.8,
        _max_ally: int = 5000,
        _max_enemy: int = 20000,
    ) -> None:
        self.all_ally = ally_villages
        self.all_enemy = enemy_villages
        self.min_radius = min_radius
        self.max_radius = max_radius
        self.precision = _precision
        self.max_ally = _max_ally
        self.max_enemy = _max_enemy
        self.neighbors = 3

    def result(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Method is used to run algorithm.

        Returns
        -------
        Tuple of np.ndarray is returned:

        (below `x` is minimum of distances to every enemy village)

        `front_lst` : array with coords `x <= min_radius`

        `back_lst` : array with coords `min_radius < x < max_radius`

        NOTES
        -----
        - Step 1: Take some part of the ally_villages and calculate `X_TRAIN` with
        precise status calculated
        - Step 2: Create KNN model based on `X_TRAIN` data
        - Step 3: Predict the rest of ally_villages with given model
        - Step 4: Return front and back villages, we do not care much about very far ones
        """
        LENGTH: int = self.all_ally.size
        LENGTH = min(max(1000, floor(self.precision * (LENGTH / 2))), self.max_ally)

        X_TRAIN: np.ndarray = self.all_ally[:LENGTH]
        X_PREDICT: np.ndarray = self.all_ally[LENGTH:]

        X_TRAIN, Y_TRAIN = self.use_cdist(
            ally_villages=X_TRAIN,
            enemy_villages=self.all_enemy[
                : min(
                    max(0.4, floor(self.all_enemy.size * self.precision)),
                    self.max_enemy,
                )
            ],
            min_radius=self.min_radius,
            max_radius=self.max_radius,
        )
        if X_PREDICT.size > 0:
            K = self.neighbors
            model = KNeighborsClassifier(n_neighbors=K)
            model.fit(X_TRAIN, Y_TRAIN)
            Y_KNN_PREDICTED = model.predict(X_PREDICT)
        else:
            Y_KNN_PREDICTED = np.array([])

        front_lst: List[np.ndarray] = []
        back_lst: List[np.ndarray] = []

        np_coords: np.ndarray
        status: Literal[0, 1, 2]
        for np_coords, status in zip(
            self.all_ally, np.append(Y_TRAIN, Y_KNN_PREDICTED)
        ):
            if status == 0:
                front_lst.append(np_coords)
            elif status == 1:
                back_lst.append(np_coords)
        return np.array(front_lst), np.array(back_lst)

    def use_cdist(
        self,
        ally_villages: np.ndarray,
        enemy_villages: np.ndarray,
        min_radius: float,
        max_radius: float,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Adds precise states to given ally_villages, calculating all possible distances
        to enemy_villages using `cdist`

        Returns
        -------
        Tuple of np.ndarray is returned:

        `lst` : array with coords

        `status` : array with three diffrent statuses where `x` is minimum of distances:

            - `0` when village is close to enemy, `x <= min_radius`
            - `1` when village is in welcome position, `min_radius < x < max_radius`
            - `2` when village is too far from enemy, `x >= max_radius`
        """

        lst: List[np.ndarray] = list()
        status: List[Literal[0, 1, 2]] = list()

        # We use sq of distance
        C = cdist(ally_villages, enemy_villages, "sqeuclidean")
        sq_min_radius: float = min_radius**2
        sq_max_radius: float = max_radius**2

        for point, i in zip(ally_villages, C):
            x: np.float64 = np.amin(i)
            lst.append(point)
            if x <= sq_min_radius:
                status.append(0)
            elif x >= sq_max_radius:
                status.append(2)
            else:
                status.append(1)

        return (
            np.array(lst),
            np.array(status),
        )



from math import floor
from typing import List, Tuple

import numpy as np
from scipy.spatial.distance import cdist


class CDistBrute:
    def __init__(
        self,
        ally_villages: np.ndarray,
        enemy_villages: np.ndarray,
        min_radius: float,
        max_radius: float,
        _precision: float = 1,
    ) -> None:
        self.all_ally = ally_villages
        self.all_enemy = enemy_villages
        self.min_radius = min_radius
        self.max_radius = max_radius
        self.precision = _precision

    def result(self, batch_size: int = 18) -> Tuple[np.ndarray, np.ndarray]:
        """
        Brute force. This is detailed version of result method.

        Returns
        -------
        Tuple of np.ndarray is returned:

        (below `x` is minimum of distances to every enemy village)

        `front_lst` : array with coords `x <= min_radius`

        `back_lst` : array with coords `min_radius < x < max_radius`

        NOTES
        -----
        For given in class init enemy villages `N` and ally villages `M`
        - Step 1: Iterate over the `M` in `i` batches `k` of size `batch_size` where `i` is ceil from `len(M)/batch_size`
        - Step 2: For every batch k1, k2 calculate matrix of sq euclidean distances k x N allies from enemies
        - Step 3: Append villages in to front, back list
        """
        triple_result = self.triple_result(batch_size)
        return triple_result[0], triple_result[1]

    def triple_result(
        self, batch_size: int = 18
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Brute force. This is detailed version of result method.

        Returns
        -------
        Tuple of np.ndarray is returned:

        (below `x` is minimum of distances to every enemy village)

        `front_lst` : array with coords `x <= min_radius`

        `back_lst` : array with coords `min_radius < x < max_radius`

        `away_lst` : array with coords `min_radius >= max_radius`

        NOTES
        -----
        For given in class init enemy villages `N` and ally villages `M`
        - Step 1: Iterate over the `M` in `i` batches `k` of size `batch_size` where `i` is ceil from `len(M)/batch_size`
        - Step 2: For every batch k1, k2 calculate matrix of sq euclidean distances k x N allies from enemies
        - Step 3: Append villages in to front, back or away list
        """

        sq_min_radius: float = self.min_radius**2
        sq_max_radius: float = self.max_radius**2
        choosen_enemy = self.all_enemy[: floor(self.all_enemy.size * self.precision)]

        total_front: List[np.ndarray] = []
        total_back: List[np.ndarray] = []
        total_away: List[np.ndarray] = []

        for i in range(int(np.ceil(self.all_ally.size / batch_size))):
            batch_ally = self.all_ally[i * batch_size : (i + 1) * batch_size]
            C = (cdist(np.array(batch_ally), choosen_enemy, "sqeuclidean")).min(axis=1)
            for index, x in np.ndenumerate(C):
                if x <= sq_min_radius:
                    total_front.append(batch_ally[index])
                elif x < sq_max_radius:
                    total_back.append(batch_ally[index])
                else:
                    total_away.append(batch_ally[index])

            # below implementation is also possible, but... slower
            # mask_front = C <= sq_min_radius
            # mask_back = np.where((C > sq_min_radius) & (C < sq_max_radius), True, False)
            # mask_away = C >= sq_max_radius
            # for array in batch_ally[mask_front]:
            #     total_front.append(array)
            # for array in batch_ally[mask_back]:
            #     total_back.append(array)
            # for array in batch_ally[mask_away]:
            #     total_away.append(array)

        return (np.array(total_front), np.array(total_back), np.array(total_away))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tw_complex']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.23.4,<2.0.0', 'scikit-learn>=1.1.3,<2.0.0', 'scipy>=1.9.3,<2.0.0']

setup_kwargs = {
    'name': 'tw-complex',
    'version': '1.1.0',
    'description': 'Algorithms for TW',
    'long_description': '![GitHub](https://img.shields.io/github/license/rafsaf/tw-complex)\n![PyPI](https://img.shields.io/pypi/v/tw-complex)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/tw-complex)\n\n# TW Complex\n\nRepo with algorithms to divide ally villages into front and back in TW.\n\nUnderneath it is a problem of dividing a set of 2D points **A** according to the `min_radius` and `max_radius` distances from a set of other 2D points **B**, which can be solved most simply by counting the distances from each point in the first set **A** to all points in the second set **B** one by one.\n\n- [Instalation](#instalation)\n- [Basic usage](#basic-usage)\n- [Examples](#examples-before---after)\n- [Running locally](#running-locally)\n\n# Instalation\n\n```bash\npip install tw_complex\n```\n\n# Basic usage\n\n```python\nfrom tw_complex import CDistBrute\nimport numpy as np\n\n# The code for Example 1 below\n\npoints1 = np.random.rand(10000, 2) + [2, 0]\npoints2 = np.random.rand(15000, 2)\nmin_radius = 1.4\nmax_radius = 2\n\nprecise_front, precise_back = CDistBrute(\n    ally_villages=points1,\n    enemy_villages=points2,\n    min_radius=min_radius,\n    max_radius=max_radius,\n).result()\n\n```\n\n# Examples (before -> after)\n\n### Example 1\n\n```bash\nAlly: 10000 points\nEnemy: 15000 points\nmin_radius: 1.4\nmax_radius: 2\n```\n\n![example1](https://raw.githubusercontent.com/rafsaf/tw-complex/main/images/Figure_1.png)\n\n### Example 2\n\n```bash\nAlly: 2500 points\nEnemy: 6000 points\nmin_radius: 4\nmax_radius: 10\n```\n\n![example2](https://raw.githubusercontent.com/rafsaf/tw-complex/main/images/Figure_2.png)\n\n### Example 3\n\n```bash\nAlly: 20000 points\nEnemy: 20000 points\nmin_radius: 20\nmax_radius: 60\n```\n\n![example3](https://raw.githubusercontent.com/rafsaf/tw-complex/main/images/Figure_3.png)\n\n### Example 4\n\n```bash\nAlly: 20000 points\nEnemy: 20000 points\nmin_radius: 10\nmax_radius: 120\n```\n\n![example4](https://raw.githubusercontent.com/rafsaf/tw-complex/main/images/Figure_4.png)\n\n# Running locally\n\nYou will need to have [poetry](https://python-poetry.org/) installed.\n\n```\ngit clone https://github.com/rafsaf/tw-complex.git\ncd tw-complex\npoetry install\n\n```\n\nCode lives in `tw-complex` folder, and you may also test algorithms running in main folder\n\n```\n# In main folder\n# eg. ~/Desktop/tw-complex\n\npytest\n```\n\nFor CDistAndKNN it looks like\n\n```python\n# tests/test_cdist.py\n\nfrom tw_complex import CDistAndKNN\nimport tests.utils as utils\n\n\ndef test_CDistAndKNN():\n    utils.run_all_tests(CDistAndKNN, "CDistAndKNN", _precision=0.8, draw=True)\n\n# Go with `draw=False` if you do not want to use pyplot to show diagrams\n```\n\nSome hardcoded tests are located in `tests/utils.py`, it uses brute force for calculating exact result, then compare it to given algorithm using basic maths. You can even compare it to brute force itself (eg. using diffrent `_precision`). For new test there should be another file in `tests/test_name_of_file_in_tw_complex_folder.py` with pretty much the same content as above.\n',
    'author': 'rafsaf',
    'author_email': 'rafal.safin12@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/rafsaf/tw-complex',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)
